#ifndef HASH_H
#define HASH_H
/*
 * Forward declaration of HashTable.  Needed by some C++ compilers
 * to prevent errors when the forward reference to HashTable is
 * encountered in the HashEntry structure.
 */

#ifdef __cplusplus
struct HashTable;
#endif


/*
 * Structure definition for an entry in a hash table.  No-one outside
 * should access any of these fields directly;  use the macros
 * defined below.
 */
typedef void * ClientData;

typedef struct HashEntry {
    struct HashEntry *nextPtr;		/* Pointer to next entry in this
					 * hash bucket, or NULL for end of
					 * chain. */
    struct HashTable *tablePtr;		/* Pointer to table containing entry. */
    struct HashEntry **bucketPtr;	/* Pointer to bucket that points to
					 * first entry in this entry's chain:
					 * used for deleting the entry. */
    ClientData clientData;		/* Application stores something here
					 * with SetHashValue. */
    union {				/* Key has one of these forms: */
	int oneWordValue;		/* One-word value for key. */
	int words[1];			/* Multiple integer words for key.
					 * The actual size will be as large
					 * as necessary for this table's
					 * keys. */
	char string[4];			/* String for key.  The actual size
					 * will be as large as needed to hold
					 * the key. */
    } key;				/* MUST BE LAST FIELD IN RECORD!! */
} HashEntry;

/*
 * Structure definition for a hash table.  Must be in tcl.h so clients
 * can allocate space for these structures, but clients should never
 * access any fields in this structure.
 */

#define SMALL_HASH_TABLE 65536
#define HASH_RATIO 0.5   /* should between 0 and 1 */
int avg_hash_size;
typedef struct HashTable {
    HashEntry **buckets;		/* Pointer to bucket array.  Each
					 * element points to first entry in
					 * bucket's hash chain, or NULL. */
    HashEntry *staticBuckets[SMALL_HASH_TABLE];
					/* Bucket array used for small tables
					 * (to avoid mallocs and frees). */
    int numBuckets;			/* Total number of buckets allocated
					 * at **bucketPtr. */
    int numEntries;			/* Total number of entries present
					 * in table. */
    int rebuildSize;			/* Enlarge table when numEntries gets
					 * to be this large. */
    int downShift;			/* Shift count used in hashing
					 * function.  Designed to use high-
					 * order bits of randomized keys. */
    int mask;				/* Mask value used in hashing
					 * function. */
    int keyType;			/* Type of keys used in this table. 
					 * It's either STRING_KEYS,
					 * ONE_WORD_KEYS, or an integer
					 * giving the number of ints that
                                         * is the size of the key.
					 */
    HashEntry *(*findProc) (struct HashTable *tablePtr,
	    char *key);
    HashEntry *(*createProc) (struct HashTable *tablePtr,
	    char *key, int *newPtr);
} HashTable;

/*
 * Structure definition for information used to keep track of searches
 * through hash tables:
 */

typedef struct HashSearch {
    HashTable *tablePtr;		/* Table being searched. */
    int nextIndex;			/* Index of next bucket to be
					 * enumerated after present one. */
    HashEntry *nextEntryPtr;	/* Next entry to be enumerated in the
					 * the current bucket. */
} HashSearch;

/*
 * Acceptable key types for hash tables:
 */

#define STRING_KEYS		0
#define ONE_WORD_KEYS		1

/*
 * Macros for clients to use to access fields of hash entries:
 */

#define GetHashValue(h) ((h)->clientData)
#define SetHashValue(h, value) ((h)->clientData = (ClientData) (value))
#define GetHashKey(tablePtr, h) \
    ((char *) (((tablePtr)->keyType == ONE_WORD_KEYS) ? (h)->key.oneWordValue \
						: (h)->key.string))

/*
 * Macros to use for clients to use to invoke find and create procedures
 * for hash tables:
 */

#define FindHashEntry(tablePtr, key) \
	(*((tablePtr)->findProc))(tablePtr, key)
#define CreateHashEntry(tablePtr, key, newPtr) \
	(*((tablePtr)->createProc))(tablePtr, key, newPtr)
#define GetHashSize(tablePtr) ((tablePtr)->numEntries) 
extern HashEntry * FirstHashEntry(HashTable*, HashSearch*);
extern HashEntry * NextHashEntry(HashSearch*);
extern void DeleteHashEntry(HashEntry *);
extern void InitHashTable(HashTable *, int);
extern void DelInitHashTable(HashTable *, int);
extern void DeleteHashTable(HashTable*);
extern char *HashStats(HashTable*);

#endif
