#ifndef CAP_LOCK_H
#define CAP_LOCK_H
#ifdef __KERNEL__
#include <linux/config.h>

/*
 *  * 'Big Reader' read-write spinlocks.
 *  *
 *  * super-fast read/write locks, with write-side penalty. The point
 *  * is to have a per-CPU read/write lock. Readers lock their CPU-local
 *  * readlock, writers must lock all locks to get write access. These
 *  * CPU-read-write locks are semantically identical to normal rwlocks.
 *  * Memory usage is higher as well. (NR_CPUS*L1_CACHE_BYTES bytes)
 *  *
 *  * The most important feature is that these spinlocks do not cause
 *  * cacheline ping-pong in the 'most readonly data' case.
 *  *
 *  * Copyright 2000, Ingo Molnar <mingo@redhat.com>
 *  *
 *  * Registry idea and naming [ crutial! :-) ] by:
 *  *
 *  *                 David S. Miller <davem@redhat.com>
 *  *
 *  * David has an implementation that doesnt use atomic operations in
 *  * the read branch via memory ordering tricks - i guess we need to
 *  * split this up into a per-arch thing? The atomicity issue is a
 *                     
 *  The atomic op version overhead is indeed a big deal on
 *  * load-locked/store-conditional cpus (ALPHA/MIPS/PPC) and
 *  * compare-and-swap cpus (Sparc64).  So we control which
 *  * implementation to use with a __BRLOCK_USE_ATOMICS define. -DaveM
 *  */


enum caplock_indices {
	CAP_CAP_LOCK,
/* You may add your item here */	
	CAP_END
};

#ifdef CONFIG_SMP

#include <linux/cache.h>
#include <linux/spinlock.h>

#if defined(__i386__) || defined(__ia64__)
#define __CAPLOCK_USE_ATOMICS
#else
#undef __CAPLOCK_USE_ATOMICS
#endif

#ifdef __CAPLOCK_USE_ATOMICS
typedef rwlock_t	caplock_read_lock_t;
#else
typedef unsigned int	caplock_read_lock_t;
#endif

#define __CAP_IDX_MAX \
	(((sizeof(caplock_read_lock_t)*CAP_END + SMP_CACHE_BYTES-1) & ~(SMP_CACHE_BYTES-1)) / sizeof(caplock_read_lock_t))

extern caplock_read_lock_t __caplock_array[NR_CPUS][__CAP_IDX_MAX];

#ifndef __CAPLOCK_USE_ATOMICS
struct cap_wrlock {
	spinlock_t lock;
} __attribute__ ((__aligned__(SMP_CACHE_BYTES)));

extern struct cap_wrlock __cap_write_locks[__CAP_IDX_MAX];
#endif

extern void __br_lock_usage_bug (void);
#define __cap_lock_usage_bug()		__br_lock_usage_bug()

#ifdef __CAPLOCK_USE_ATOMICS

static inline void cap_read_lock (enum caplock_indices idx)
{
	read_lock(&__caplock_array[smp_processor_id()][idx]);
}

static inline void cap_read_unlock (enum caplock_indices idx)
{
	read_unlock(&__caplock_array[smp_processor_id()][idx]);
}

#else /* ! __CAPLOCK_USE_ATOMICS */
static inline void cap_read_lock (enum caplock_indices idx)
{
	unsigned int *ctr;
	spinlock_t *lock;

	ctr = &__caplock_array[smp_processor_id()][idx];
	lock = &__cap_write_locks[idx].lock;
again:
	(*ctr)++;
	mb();
	if (spin_is_locked(lock)) {
		(*ctr)--;
		wmb(); 
		while (spin_is_locked(lock))
			barrier();
		goto again;
	}
}

static inline void cap_read_unlock (enum caplock_indices idx)
{
	unsigned int *ctr;

	ctr = &__caplock_array[smp_processor_id()][idx];

	wmb();
	(*ctr)--;
}

#endif /* __CAPLOCK_USE_ATOMICS */

extern void FASTCALL(__cap_write_lock (enum caplock_indices idx));
extern void FASTCALL(__cap_write_unlock (enum caplock_indices idx));

static inline void cap_write_lock (enum caplock_indices idx)
{
	__cap_write_lock(idx);
}

static inline void cap_write_unlock (enum caplock_indices idx)
{
	__cap_write_unlock(idx);
}

#else
# define cap_read_lock(idx)	((void)(idx))
# define cap_read_unlock(idx)	((void)(idx))
# define cap_write_lock(idx)	((void)(idx))
# define cap_write_unlock(idx)	((void)(idx))
#endif

extern int register_caplock(unsigned long data);
extern int unregister_caplock(unsigned long data);

#define cap_read_lock_irqsave(idx, flags) \
	do { local_irq_save(flags); cap_read_lock(idx); } while (0)
#define cap_read_lock_irq(idx) \
	do { local_irq_disable(); cap_read_lock(idx); } while (0)
#define cap_read_lock_bh(idx) \
	do { local_bh_disable(); cap_read_lock(idx); } while (0)
#define cap_write_lock_irqsave(idx, flags) \
	do { local_irq_save(flags); cap_write_lock(idx); } while (0)
#define cap_write_lock_irq(idx) \
	do { local_irq_disable(); cap_write_lock(idx); } while (0)
#define cap_write_lock_bh(idx) \
	do { local_bh_disable(); cap_write_lock(idx); } while (0)
#define cap_read_unlock_irqrestore(idx, flags) \
	do { cap_read_unlock(irx); local_irq_restore(flags); } while (0)
#define cap_read_unlock_irq(idx) \
	do { cap_read_unlock(idx); local_irq_enable(); } while (0)
#define cap_read_unlock_bh(idx) \
	do { cap_read_unlock(idx); local_bh_enable(); } while (0)
#define cap_write_unlock_irqrestore(idx, flags) \
	do { cap_write_unlock(irx); local_irq_restore(flags); } while (0)
#define cap_write_unlock_irq(idx) \
	do { cap_write_unlock(idx); local_irq_enable(); } while (0)
#define cap_write_unlock_bh(idx) \
	do { cap_write_unlock(idx); local_bh_enable(); } while (0)

#endif	
#endif
